searchData={"items":[{"type":"module","title":"SeedFactory","doc":"A toolkit for test data generation.\n\nThe main idea of `SeedFactory` is to generate data in tests according to your application business logic (read as\ncontext functions if you use [Phoenix Contexts](https://hexdocs.pm/phoenix/contexts.html)) whenever it is possible\nand avoid direct inserts to the database (as opposed to `ex_machina`).\nThis approach allows you to minimize testing of invalid states as you're not forced to keep complex database structure in your head in order to prepare test data.\nThe library is completely agnostic to the database toolkit.\n\n**Context**, **entities** and **commands** are the core concepts of the library.\n\nContext is a `t:map/0` which can be populated with entities using commands.\n\nThe schema with instructions on how commands modify context is described using DSL with the help of `SeedFactory.Schema` module.\n\nCommands can be used to:\n* produce entity (put new data in the context)\n* update entity (replace the existing entity in the context)\n* delete entity (remove the entity from the context)\n\nWhen a command is executed, produced entities are assigned to the context using the name of the entity as a key.\nA command has params with instructions on how to generate arguments for a resolver if they are not passed explicitly with `exec/3` function.\nThe instruction can be specified using one of these options:\n* `:value` - any term. This option is used by default with the value of `nil`.\n* `:generate` - a zero-arity function for generating data.\n* `:entity` - an atom which points to an entity which should be taken from the context. If a required entity cannot\nbe found in a context, then `SeedFactory` automatically executes a corresponding command which produces the entity.\n\nEntities can have traits.\nThink about them as labels which are assigned to produced/updated entities when specific commands with specific arguments are executed.\n\nLet's take a look at an example of a simple schema.\n```elixir\ndefmodule MyApp.SeedFactorySchema do\n  use SeedFactory.Schema\n\n  command :create_company do\n    param :name, generate: &Faker.Company.name/0\n\n    resolve(fn args ->\n      with {:ok, company} <- MyApp.Companies.create_company(args) do\n        {:ok, %{company: company}}\n      end\n    end)\n\n    produce :company\n  end\n\n  command :create_user do\n    param :name, generate: &Faker.Person.name/0\n    param :role, value: :normal\n    param :company, entity: :company\n\n    resolve(fn args -> MyApp.Users.create_user(args.company, args.name, args.role) end)\n\n    produce :user\n    produce :profile\n  end\n\n  command :activate_user do\n    param :user, entity: :user, with_traits: [:pending]\n\n    resolve(fn args ->\n      user = MyApp.Users.activate_user!(args.user)\n\n      {:ok, %{user: user}}\n    end)\n\n    update :user\n  end\n\n  trait :pending, :user do\n    exec :create_user\n  end\n\n  trait :active, :user do\n    from :pending\n    exec :activate_user\n  end\n\n  trait :admin, :user do\n    exec :create_user, args_pattern: %{role: :admin}\n  end\n\n  trait :normal, :user do\n    exec :create_user, args_pattern: %{role: :normal}\n  end\nend\n```\nThe schema above describes how to produce 3 entities (`:company`, `:user` and `:profile`) using 2 commands (`:create_user` and `:create_company`).\nThere is a third command which only updates the `:user` entity.\nThere are 4 traits defined for the `:user` entity.\n\nTo start using the schema, put metadata about it to the context using `init/2` function:\n```elixir\ncontext = %{}\ncontext = init(context, MyApp.SeedFactorySchema)\n```\nIf you use `SeedFactory` in tests with `ExUnit`, check out `SeedFactory.Test`. This module adds initialization using `ExUnit.Callbacks.setup_all/2` callback\nand imports functions.\n\nNow, `exec/2` function can be used to execute a command:\n```elixir\ncontext = exec(context, :create_company)\n```\nThe code above will generate arguments for `:create_company` command, execute a resolver with generated arguments and put company to context using `:company` key.\n`exec/3` can be used if you want to specify parameters explicitly:\n```elixir\ncontext = exec(context, :create_company, name: \"GitHub\")\n```\n\nBecause exec function returns `t:context/0`, it is convenient to chain `exec` calls with the pipe operator:\n```elixir\ncontext =\n  context\n  |> exec(:create_company)\n  |> exec(:create_user, name: \"John Doe\")\n```\nIn order to get a value for the `:company` parameter of the `:create_user` command, the corresponding entity was taken from the context.\nHowever, it is not necessary to do so, as `SeedFactory` can automatically execute commands which produce dependent entities.\nThe code above has the same effect as a single call to `:create_user` command:\n```elixir\ncontext = exec(context, :create_user, name: \"John Doe\")\n```\n\nIf you're not interested in explicit providing of parameters to commands, then you can use `produce/2` function to produce\nrequested entities with automatic execution of all dependent commands:\n```elixir\ncontext = produce(context, :user)\n```\nEven though `:user` is the only entity specified explicitly, `context` will have 3 new keys: `:company`, `:user` and `:profile`.\n\n> #### Tip {: .tip}\n> It is recommended to explicitly specify all entities in which you're insterested:\n> ```elixir\n> # good\n> %{user: user} = produce(contex, :user)\n>\n> # good\n> %{user: user, profile: profile} = produce(contex, [:user, :profile])\n>\n> # good\n> %{user: user, company: company} = produce(contex, [:user, :company])\n>\n> # bad\n> %{user: user, profile: profile, company: company} = produce(contex, :user)\n> ```\n\n`exec/3` fails if produced entities are already present in the context.\nIt is possible to rebind entities in order to assign them to the context with different names:\n```elixir\ncontext =\n  context\n  |> rebind([user: :user1, profile: :profile1], &exec(&1, :create_user))\n  |> rebind([user: :user2, profile: :profile2], &exec(&1, :create_user))\n```\nThe snippet above puts the following keys to the context: `:company`, `:user1`, `:profile1`, `:user2` and `:profile2`.\nThe `:company` is shared in this case, so two users have different profiles and belong to the same company.\nA shorter counterpart using `produce/2` is the following:\n```elixir\ncontext =\n  context\n  |> produce(user: :user1, profile: :profile1)\n  |> produce(user: :user2, profile: :profile2)\n```\n\nAs was pointed out previously, traits are assigned to entities when commands produce/update them.\n`SeedFactory` does this automatically by tracking commands and arguments.\nYou can inspect `__seed_factory_meta__` key in the context to review currently assigned traits:\n\n```elixir\ncontext |> exec(:create_user) |> IO.inspect()\n# %{\n# __seed_factory_meta__: #SeedFactory.Meta<\n#   current_traits: %{user: [:normal, :pending]},\n#   ...\n# >,\n# ...\n# }\n\ncontext |> exec(:create_user, role: :admin) |> exec(:activate_user) |> IO.inspect()\n# %{\n# __seed_factory_meta__: #SeedFactory.Meta<\n#   current_traits: %{user: [:admin, :active]},\n#   ...\n# >,\n# ...\n# }\n```\n\nThe same result can be achieved by passing traits using `produce/2`:\n```elixir\nproduce(context, user: [:admin, :active])\n```\n\nIf you want to specify traits and assign an entity to the context with the different name, then use `:as` option:\n```elixir\n%{active_admin: active_admin} = produce(context, user: [:admin, :active, as: :active_admin])\n```","ref":"SeedFactory.html"},{"type":"function","title":"SeedFactory.exec/3","doc":"Executes a command and updates the `context` according to the schema.","ref":"SeedFactory.html#exec/3"},{"type":"function","title":"Example - SeedFactory.exec/3","doc":"iex> context = %{}\n    ...> context = init(context, MySeedFactorySchema)\n    ...> context = exec(context, :create_user, first_name: \"John\", last_name: \"Doe\")\n    ...> Map.take(context.user, [:first_name, :last_name])\n    %{first_name: \"John\", last_name: \"Doe\"}","ref":"SeedFactory.html#exec/3-example"},{"type":"function","title":"SeedFactory.init/2","doc":"Puts metadata about `schema` to `context`, so `context` becomes usable by other functions from this module.","ref":"SeedFactory.html#init/2"},{"type":"function","title":"Example - SeedFactory.init/2","doc":"iex> context = %{}\n    ...> init(context, MySeedFactorySchema)\n    %{__seed_factory_meta__: #SeedFactory.Meta<...>}","ref":"SeedFactory.html#init/2-example"},{"type":"function","title":"SeedFactory.pre_exec/3","doc":"Creates dependent entities needed for command execution.\n\nThis is useful, when you want to execute the command multiple times reusing input entities.","ref":"SeedFactory.html#pre_exec/3"},{"type":"function","title":"Example - SeedFactory.pre_exec/3","doc":"# Function :create_user produces multiple entities (:user and :profile), so if you want to\n    # produce multiple users using a sequence of `exec` calls, you have to write this:\n    %{user1: user1, user2: user2} =\n      context\n      |> rebind([user: :user1, profile: :profile1], &exec(&1, :create_user, role: :admin))\n      |> rebind([user: :user2, profile: :profile2], &exec(&1, :create_user, role: :admin))\n\n    # The code above is a bit wordy in a case when all we need are :user entities. We have to write\n    # rebinding for :profile even though we are't interested in it.\n\n    # A less wordy alternative is:\n    context = produce(context, list_of_all_dependencies_with_their_traits)\n    %{user: user1} = exec(context, :create_user, role: :admin)\n    %{user: user2} = exec(context, :create_user, role: :admin)\n\n    # However, you have to explicitly enumerate all the dependencies.\n    # It can be more compact with `pre_exec` function:\n    context = pre_exec(context, :create_user)\n    %{user: user1} = exec(context, :create_user, role: :admin)\n    %{user: user2} = exec(context, :create_user, role: :admin)","ref":"SeedFactory.html#pre_exec/3-example"},{"type":"function","title":"SeedFactory.pre_produce/2","doc":"Produces dependencies needed for specified entities.\n\nSee `pre_exec/3` for problems it solves.","ref":"SeedFactory.html#pre_produce/2"},{"type":"function","title":"Example - SeedFactory.pre_produce/2","doc":"# pre_produce produces a company and puts it into context,\n    # so produced user1 and user2 will belong to the same company\n    context = pre_produce(context, :user)\n    %{user: user1} = produce(context, :user)\n    %{user: user2} = produce(context, :user)","ref":"SeedFactory.html#pre_produce/2-example"},{"type":"function","title":"SeedFactory.produce/2","doc":"Produces entities if they don't exist in the context.\n\nThe order of specified entities doesn't matter.\n\nIt invokes a series of commands to produce entities with specified traits.\nIf the same entity can be produced by multiple commands, then the first declared command is used by default.\nIn order to produce the entity using the rest of the commands use traits or `exec/3` explicitly.","ref":"SeedFactory.html#produce/2"},{"type":"function","title":"Examples - SeedFactory.produce/2","doc":"# specify a single entity\n    %{user: _} = produce(context, :user)\n\n    # specify a list of entitities\n    %{user: _, company: _} = produce(context, [:user, :company])\n\n    # rebind :user entity as :user1\n    %{user1: _} = produce(context, user: :user1)\n\n    # specify traits\n    %{user: _} = produce(context, user: [:active, :admin])\n    %{user: _} = context |> produce(user: [:pending, :admin]) |> produce(user: [:active])\n\n    # specify traits and :as option\n    %{user1: _} = produce(context, user: [:active, :admin, as: :user1])","ref":"SeedFactory.html#produce/2-examples"},{"type":"function","title":"SeedFactory.rebind/3","doc":"Changes default behaviour of entities assignment.\n\nModifies a context, so commands engage with entities in the context using provided names.\nIf a command requires/produces/updates/deletes an entity, the provided name will be used instead of the entity name to get value from the context or modify it.\n\nIt helps to produce the same entity multiple times and assign it to the context with different names.","ref":"SeedFactory.html#rebind/3"},{"type":"function","title":"Example - SeedFactory.rebind/3","doc":"# Let's assume that `:product` entity requires `:company` entity.\n    # In this example we create 2 companies, `:product1` will be linked to `:company1`.\n    %{company1: _, company2: _, product1: _} =\n      context\n      |> rebind([company: :company1, product: :product1], fn context ->\n        context\n        |> exec(:create_company, name: \"GitHub\")\n        |> produce(:product)\n      end)\n      |> rebind([company: :company2], fn context ->\n        exec(context, :create_company, name: \"Microsoft\")\n      end)","ref":"SeedFactory.html#rebind/3-example"},{"type":"type","title":"SeedFactory.context/0","doc":"","ref":"SeedFactory.html#t:context/0"},{"type":"type","title":"SeedFactory.entity_name/0","doc":"","ref":"SeedFactory.html#t:entity_name/0"},{"type":"type","title":"SeedFactory.rebinding_rule/0","doc":"","ref":"SeedFactory.html#t:rebinding_rule/0"},{"type":"module","title":"SeedFactory.Schema","doc":"A schema describes how commands modify context\n\nThis module provides a DSL for defining schemas that describe how entities should be\ncreated, updated, or deleted within a context using the `SeedFactory` library.\n\nIn order to use the DSL, create a schema module with the following content:\n\n    defmodule MyApp.SeedFactorySchema do\n      use SeedFactory.Schema\n    end","ref":"SeedFactory.Schema.html"},{"type":"module","title":"Command Definition - SeedFactory.Schema","doc":"Command is the first thing that should be defined in the schema.\nTo define a command, use the `command` macro followed by the command name. Inside the command block,\nyou can define input parameters, a resolution, and produce, update, and delete directives.\n\n```elixir\ncommand :create_user do\n  # Parameters, resolution, produce, update, and delete directives\nend\n```","ref":"SeedFactory.Schema.html#module-command-definition"},{"type":"module","title":"Parameters - SeedFactory.Schema","doc":"Parameters define the inputs required for the command's resolver function and how default values should be generated.\nParameter can be defined using the `param` macro and can have an arbitrary level of nesting.","ref":"SeedFactory.Schema.html#module-parameters"},{"type":"module","title":"Options - SeedFactory.Schema","doc":"* `:value` - a static default value. Applied by default as `value: nil`.\n* `:generate` - an anonymous function that generates data.\n* `:entity` - refers to an entity within the context.\n* `:with_traits` - a list of atoms with trait names. Can be applied only if `:entity` option is present.\n* `:map` - an anonymous function that allows mapping an entity to another value. Can be applied only if `:entity` option is present.\n\n```elixir\nparam :address do\n  param :city, value: \"Lemberg\"\n  param :street, generate: &random_street/0\nend\n\nparam :paid_by, entity: :user, with_traits: [:active]\nparam :office_id, entity: :office, map: & &1.id\n\nparam :github_username\n# the line above is equivalent to\n# param :github_username, value: nil\n```","ref":"SeedFactory.Schema.html#module-options"},{"type":"module","title":"Resolution - SeedFactory.Schema","doc":"The resolution defines the logic to be executed when the command is invoked. It is implemented as a resolver function inside the `resolve` macro.\nThe resolver function is an anonymous function that takes `args` as its parameter.\n\nIt should return `{:ok, map}`, where map keys are atoms and values represent entities.\nThe atom keys will be used by the `:from` option in `produce` and `update` directives.\n\n`{:error, reason}` will abort the command execution by raising an exception.\n\n```elixir\nresolve(fn args ->\n  user = MyApp.insert_user!(args)\n  {:ok, %{user: user}}\nend)\n```","ref":"SeedFactory.Schema.html#module-resolution"},{"type":"module","title":"Producing Entities - SeedFactory.Schema","doc":"The `produce` directive specifies that the command will put a new entity to the context. It takes two arguments: the name of the entity being produced and options.","ref":"SeedFactory.Schema.html#module-producing-entities"},{"type":"module","title":"Options - SeedFactory.Schema","doc":"* `:from` - an atom that specifies the key of the map returned by the resolver. Defaults to specified entity name (first argument)\n\n```elixir\nproduce :entity_name, from: :source_key\n```\n\n```elixir\nresolve(fn args ->\n  {user, profile} = MyApp.register_user!(args)\n  {:ok, %{user: user, profile: profile}}\nend)\n\nproduce :user\nproduce :user_profile, from: :profile\n```","ref":"SeedFactory.Schema.html#module-options"},{"type":"module","title":"Updating Entities - SeedFactory.Schema","doc":"The `update` directive modifies an existing entity within the context. It takes two arguments: the name of the entity being updated and options.","ref":"SeedFactory.Schema.html#module-updating-entities"},{"type":"module","title":"Options - SeedFactory.Schema","doc":"* `:from` - an atom that specifies the key of the map returned by the resolver. Defaults to specified entity name (first argument)\n\n```elixir\nupdate :entity_name, from: :source_key\n```\n\n```elixir\nresolve(fn args ->\n  {user, profile} = MyApp.update_user!(args)\n  {:ok, %{user: user, profile: profile}}\nend)\n\nupdate :user\nupdate :user_profile, from: :profile\n```","ref":"SeedFactory.Schema.html#module-options"},{"type":"module","title":"Deleting Entities - SeedFactory.Schema","doc":"The `delete` directive removes an entity from the context.\n\n```elixir\nresolve(fn args ->\n  MyApp.delete_user!(args.user_id)\n  {:ok, %{}}\nend)\n\ndelete :user\n```","ref":"SeedFactory.Schema.html#module-deleting-entities"},{"type":"module","title":"Traits - SeedFactory.Schema","doc":"The `trait` directive declares a trait for an entity.\nThe first argument is trait name, the second is entity name.\n\nA trait must contain an `exec` directive with the name of the command. Execution of the specified command marks\nentity with the trait.","ref":"SeedFactory.Schema.html#module-traits"},{"type":"module","title":"Options - SeedFactory.Schema","doc":"* `:from` - an atom or a list of atoms that point to the traits that should be replaced with the new one. This is useful for different kinds of\nstatus transitions.\n\n\n```elixir\ntrait :pending, :user do\n  exec :create_user\nend\n\ntrait :active, :user do\n  from :pending\n  exec :activate_user\nend\n\ntrait :suspended, :user do\n  from [:pending, :active]\n  exec :suspend_user\nend\n\n# execute :create_user command\nproduce(ctx, user: [:pending])\n\n# execute :create_user -> :activate_user\nproduce(ctx, user: [:active])\n\n# execute :create_user -> :suspend_user\nproduce(ctx, user: [:suspended])\n\n# execute :create_user -> :activate_user -> :suspend_user\nctx |> produce(user: [:active]) |> produce(user: [:suspended])\n```","ref":"SeedFactory.Schema.html#module-options"},{"type":"module","title":"Exec step - SeedFactory.Schema","doc":"The `exec` directive is required when declaring a trait and is used for specifying what should be executed in order\nto mark entity with the trait","ref":"SeedFactory.Schema.html#module-exec-step"},{"type":"module","title":"Options - SeedFactory.Schema","doc":"* `:args_match` - a function which accepts command args and must return a boolean. Must be used with `:generate_args` option.\nIf the function returns `true`, then the entity will be marked with the trait.\n* `:generate_args` - a function which generates a map with args. Must be used with `:args_match` option. The function generates args which satisfy\nvalidation in `:args_match` option and is used when the entity is requested with the trait.\n* `:args_pattern` - a map with args. This option option is less verbose (and limited in functionality) alternative to the combination of `:args_match` and `:generate_args` options.\nIf specified, then entity will be marked with the trait only when command args match the pattern. Also, the pattern will be used as a replacement to `:generate_args` invocation.\n\n```elixir\n# all three instructions below are equal\nexec :create_user\nexec :create_user, args_pattern: %{}\nexec :create_user, generate_args: fn -> %{} end, args_match: fn _args -> true end\n```\n\n```elixir\ntrait :admin, :user do\n  exec :create_user, args_pattern: %{role: :admin}\nend\n\ntrait :normal, :user do\n  exec :create_user, args_pattern: %{role: :normal}\nend\n\n# the same using the combination of `:args_match` and `:generate_args`\ntrait :admin, :user do\n  exec :create_user do\n    generate_args(fn -> %{role: :admin} end)\n    args_match(&match?(%{role: :admin}, &1)\n  end\nend\n\ntrait :normal, :user do\n  exec :create_user do\n    generate_args(fn -> %{role: :normal} end)\n    args_match(&match?(%{role: :normal}, &1))\n  end\nend\n```\n\n```elixir\n# an example which shows what is possible with `:args_match` + `:generate_args`\n# but not with `:args_pattern`\n\ntrait :not_expired, :project do\n  exec :publish_project do\n    args_match(fn args -> Date.compare(Date.utc_today(), args.expiry_date) in [:lt, :eq] end)\n\n    generate_args(fn ->\n      today = Date.utc_today()\n      %{start_date: today, expiry_date: Date.add(today, 21)}\n    end)\n  end\nend\n\ntrait :expired, :project do\n  exec :publish_project do\n    args_match(fn args -> Date.compare(Date.utc_today(), args.expiry_date) == :gt end)\n\n    generate_args(fn ->\n      today = Date.utc_today()\n      %{start_date: Date.add(today, -22), expiry_date: Date.add(today, -1)}\n    end)\n  end\nend\n```","ref":"SeedFactory.Schema.html#module-options"},{"type":"module","title":"Include schemas - SeedFactory.Schema","doc":"It is possible to include multiple schemas into a new schema in order to reuse everything that is declared in specified modules.\n\n```elixir\ndefmodule MyAppWeb.SeedFactorySchema do\n  use SeedFactory.Schema\n\n  include_schema MyApp.SeedFactorySchema\n\n  # Web-specific stuff goes here. You may need such a separation of modules\n  # if you have an umbrella project and web is a separate app\n  command :build_conn do\n    resolve(fn _ ->\n      conn =\n        Phoenix.ConnTest.build_conn()\n        |> Plug.Conn.put_private(:phoenix_endpoint, MyAppWeb.Endpoint)\n      {:ok, %{conn: conn}}\n    end)\n\n    produce :conn\n  end\n\n  command :create_user_session do\n    param :user, entity: :user, with_traits: [:active]\n    param :conn, entity: :conn, with_traits: [:unauthenticated]\n\n    resolve(fn args ->\n      {:ok, %{conn: MyAppWeb.Session.init_user_session(args.conn, args.user)}}\n    end)\n\n    update :conn\n  end\n\n  trait :unauthenticated, :conn do\n    exec :build_conn\n  end\n\n  trait :user_session, :conn do\n    from :unauthenticated\n    exec :create_user_session\n  end\nend\n```","ref":"SeedFactory.Schema.html#module-include-schemas"},{"type":"module","title":"SeedFactory.Test","doc":"A helper module for `ExUnit`.","ref":"SeedFactory.Test.html"},{"type":"module","title":"Usage - SeedFactory.Test","doc":"Add the following line to your test modules:\n```\nuse SeedFactory.Test, schema: MySeedFactorySchema\n```\nIt sets up `SeedFactory` by invoking `SeedFactory.init/2` in `ExUnit.Callbacks.setup_all/2` block and imports the following functions:\n  * `produce/1`\n  * `SeedFactory.rebind/3`\n  * `SeedFactory.produce/2`\n  * `SeedFactory.exec/2`\n  * `SeedFactory.exec/3`\n  * `SeedFactory.pre_exec/2`\n  * `SeedFactory.pre_exec/3`\n  * `SeedFactory.pre_produce/2`","ref":"SeedFactory.Test.html#module-usage"},{"type":"macro","title":"SeedFactory.Test.produce/1","doc":"A macro that implicitly passes `context` and allows usage of `SeedFactory.produce/2` outside the `test` block.\n\nBasically, it creates a `ExUnit.Callbacks.setup/2` block and calls `SeedFactory.produce/2` inside.","ref":"SeedFactory.Test.html#produce/1"},{"type":"macro","title":"Examples - SeedFactory.Test.produce/1","doc":"```elixir\nproduce :company\n\ntest \"my test\", %{company: company} do\n  assert my_function(company)\nend\n```\n\n```elixir\nproduce [:user, :project]\n\ntest \"my test\", %{user: user, project: project} do\n  assert my_function(project, user)\nend\n```\n\n```elixir\nproduce org: :org1\nproduce org: :org2\n\ntest \"my test\", %{org1: org1, org2: org2} do\n  assert my_function(org2, org1)\nend\n```","ref":"SeedFactory.Test.html#produce/1-examples"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,50,46,50]}}